from otlmow_model.BaseClasses.StringField import StringField


class URIField(StringField):
    """Een tekstwaarde die een verwijzing naar meer informatie van het element bevat volgens http://www.w3.org/2001/XMLSchema#anyURI ."""
    naam = 'AnyURI'
    objectUri = 'http://www.w3.org/2001/XMLSchema#anyURI'
    definition = 'Een tekstwaarde die een verwijzing naar meer informatie van het element bevat volgens http://www.w3.org/2001/XMLSchema#anyURI.'
    label = 'URI'
    usagenote = 'https://www.w3.org/TR/xmlschema-2/#anyURI'

    @staticmethod
    def validate(value, attribuut):
        return StringField.validate(value, attribuut)
        # TODO add URI validation

    def __str__(self):
        return StringField.__str__(self)




import inspect
import warnings
from typing import Type, Optional

from otlmow_model.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_model.Exceptions.RelationDeprecationWarning import RelationDeprecationWarning
from otlmow_model.Helpers.AssetCreator import AssetCreator


class RelationValidator:
    @staticmethod
    def is_valid_relation_instance(source: RelationInteractor, relation_instance: RelatieObject, 
                                   target: RelationInteractor) -> bool:
        """
        Verifies if a relation would be valid between a source and a target, given the instance of that relation

        :param source: the intended source for the relation
        :param relation_instance: the instance of the relation
        :param target: the intended source for the relation
        :return: 'True' if the relation would be valid, 'False' otherwise
        """
        return RelationValidator.is_valid_relation(source=source, target=target,
                                                   relation_type=type(relation_instance))

    @staticmethod
    def is_valid_relation(relation_type: Type[RelatieObject], source: RelationInteractor = None,
                          source_typeURI: Optional[str] = None, target: RelationInteractor = None,
                          target_typeURI: Optional[str] = None) -> bool:
        """
        Verifies if a relation would be valid between a source and a target, given a relation_type type. Exactly one of source or source_typeURI must not be None. Exactly one of target or target_typeURI must not be None.

        :param relation_type: the intended type of the relation
        :param source: the intended source for the relation_type
        :param source_typeURI: the typeURI of the intended source for the relation
        :param target: the intended source for the relation
        :param target_typeURI: the typeURI of the intended target for the relation
        :return: 'True' if the relation would be valid, 'False' otherwise
        """

        if source is not None and source_typeURI is not None:
            warnings.warn('both source and source_typeURI are not None. Overriding source by instantiating it',
                          RuntimeWarning)
            source = AssetCreator.dynamic_create_instance_from_uri(source_typeURI)
        if target is not None and target_typeURI is not None:
            warnings.warn('both target and target_typeURI are not None. Overriding source by instantiating it',
                          RuntimeWarning)
            target = AssetCreator.dynamic_create_instance_from_uri(source_typeURI)

        if source is not None and source_typeURI is None:
            source_typeURI = source.typeURI
        if source_typeURI is None:
            raise ValueError('Exactly one of source or source_typeURI needs to be not None.')

        if target is not None and target_typeURI is None:
            target_typeURI = target.typeURI
        if target_typeURI is None:
            raise ValueError('Exactly one of target or target_typeURI needs to be not None.')

        if 'lgc.' in source_typeURI or 'lgc.' in target_typeURI:
            return True

        if source is None:
            source = AssetCreator.dynamic_create_instance_from_uri(source_typeURI)

        if relation_type.typeURI not in source._valid_relations:
            return False

        targets = source._valid_relations[relation_type.typeURI].keys()
        if target_typeURI in targets:
            deprecated_value = source._valid_relations[relation_type.typeURI][target_typeURI]
            if deprecated_value != '':
                warnings.warn(
                    message=f'the relation_type of type {relation_type.typeURI} between assets of types '
                            f'{source_typeURI} and {target_typeURI} is deprecated since version {deprecated_value}',
                    category=RelationDeprecationWarning)
            return True

        if target is None:
            target = AssetCreator.dynamic_create_instance_from_uri(target_typeURI)
        bases = inspect.getmro(type(target))
        for base in bases:
            base_type_uri = RelationValidator._get_member(base, 'typeURI')
            if base_type_uri in targets:
                deprecated_value = source._valid_relations[relation_type.typeURI][base_type_uri]
                if deprecated_value != '':
                    warnings.warn(
                        message=f'the relation_type of type {relation_type.typeURI} between assets of types '
                                f'{source_typeURI} and {target_typeURI} is deprecated since version {deprecated_value}',
                        category=RelationDeprecationWarning)
                return True

        return False

    @staticmethod
    def _get_member(obj, name):
        return next(iter([member for _name, member in inspect.getmembers(obj) if name == _name]), None)



from typing import Union

from otlmow_model.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_model.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_model.Helpers.GenericHelper import get_titlecase_from_ns, get_ns_and_name_from_uri


class AssetCreator:
    @staticmethod
    def dynamic_create_instance_from_ns_and_name(namespace: str, class_name: str,
                                                 directory: str = 'otlmow_model.Classes') -> Union[AIMObject, RelatieObject]:
        """Loads the OTL class module and attempts to instantiate the class using the name and namespace of the class

        :param namespace: namespace of the class
        :type: str
        :param class_name: class name to instantiate
        :type: str
        :param directory: directory where the class modules are located, defaults to OTLMOW.OTLModel.Classes
        :type: str
        :return: returns an instance of class_name in the given namespace, located from directory, that inherits from AIMObject or RelatieObject
        :rtype: AIMObject, RelatieObject or None
        """

        if directory is None:
            directory = 'otlmow_model.Classes'

        if namespace is None:
            namespace = ''
        else:
            namespace = get_titlecase_from_ns(namespace)

        try:
            # TODO: check https://stackoverflow.com/questions/2724260/why-does-pythons-import-require-fromlist
            py_mod = __import__(name=f'{directory}.{namespace}.{class_name}', fromlist=f'{class_name}')
        except ModuleNotFoundError:
            return None
        class_ = getattr(py_mod, class_name)
        instance = class_()

        return instance

    @staticmethod
    def dynamic_create_instance_from_uri(class_uri: str, directory: str = None) -> Union[AIMObject, RelatieObject]:
        if directory is None:
            directory = 'otlmow_model.Classes'

        if not class_uri.startswith('https://wegenenverkeer.data.vlaanderen.be/ns'):
            raise ValueError(
                f'{class_uri} is not valid uri, it does not begin with "https://wegenenverkeer.data.vlaanderen.be/ns"')
        ns, name = get_ns_and_name_from_uri(class_uri)
        created = AssetCreator.dynamic_create_instance_from_ns_and_name(ns, name, directory=directory)
        if created is None:
            raise ValueError(f'{class_uri} is likely not valid uri, it does not result in a created instance')
        return created